@@ -34,7 +34,7 @@
 #include <unistd.h>
 #endif
 
-#include "ocfs2.h"
+#include "ocfs2/ocfs2.h"
 
 static errcode_t open_namei(ocfs2_filesys *fs, uint64_t root, uint64_t base,
 			    const char *pathname, size_t pathlen, int follow,
@@ -93,7 +93,7 @@ static errcode_t follow_link(ocfs2_filesys *fs, uint64_t root, uint64_t dir,
 	if (ret)
 		goto bail;
 
-	ret = io_read_block(fs->fs_io, blkno, 1, buffer);
+	ret = ocfs2_read_blocks(fs, blkno, 1, buffer);
 	if (ret)
 		goto bail;
 
