@@ -82,6 +82,7 @@ static AllocGroup * initialize_alloc_group(State *s, const char *name,
 					   uint64_t blkno,
 					   uint16_t chain, uint16_t cpg,
 					   uint16_t bpc);
+static void index_system_dirs(State *s, ocfs2_filesys *fs);
 static void create_lost_found_dir(State *s, ocfs2_filesys *fs);
 static void format_journals(State *s, ocfs2_filesys *fs);
 static void format_slotmap(State *s, ocfs2_filesys *fs);
@@ -98,23 +99,28 @@ static SystemFileInfo system_files[] = {
 	{ "slot_map", SFI_OTHER, 1, S_IFREG | 0644 },
 	{ "heartbeat", SFI_HEARTBEAT, 1, S_IFREG | 0644 },
 	{ "global_bitmap", SFI_CLUSTER, 1, S_IFREG | 0644 },
+	{ "aquota.user", SFI_QUOTA, 1, S_IFREG | 0644 },
+	{ "aquota.group", SFI_QUOTA, 1, S_IFREG | 0644 },
 	{ "orphan_dir:%04d", SFI_OTHER, 0, S_IFDIR | 0755 },
 	{ "extent_alloc:%04d", SFI_CHAIN, 0, S_IFREG | 0644 },
 	{ "inode_alloc:%04d", SFI_CHAIN, 0, S_IFREG | 0644 },
 	{ "journal:%04d", SFI_JOURNAL, 0, S_IFREG | 0644 },
 	{ "local_alloc:%04d", SFI_LOCAL_ALLOC, 0, S_IFREG | 0644 },
-	{ "truncate_log:%04d", SFI_TRUNCATE_LOG, 0, S_IFREG | 0644 }
+	{ "truncate_log:%04d", SFI_TRUNCATE_LOG, 0, S_IFREG | 0644 },
+	{ "aquota.user:%04d", SFI_QUOTA, 0, S_IFREG | 0644 },
+	{ "aquota.group:%04d", SFI_QUOTA, 0, S_IFREG | 0644 },
 };
 
 struct fs_type_translation {
 	const char *ft_str;
-	enum ocfs2_fs_types ft_type;
+	enum ocfs2_mkfs_types ft_type;
 };
 
-static struct fs_type_translation ocfs2_fs_types_table[] = {
-	{"datafiles", FS_DATAFILES},
-	{"mail", FS_MAIL},
-	{NULL, FS_DEFAULT},
+static struct fs_type_translation ocfs2_mkfs_types_table[] = {
+	{"datafiles", OCFS2_MKFSTYPE_DATAFILES},
+	{"mail", OCFS2_MKFSTYPE_MAIL},
+	{"vmstore", OCFS2_MKFSTYPE_VMSTORE},
+	{NULL, OCFS2_MKFSTYPE_DEFAULT},
 };
 
 enum {
@@ -226,6 +232,23 @@ static void mkfs_init_dir_trailer(State *s, DirData *dir, void *buf)
 	}
 }
 
+/* Should we skip this inode because of features enabled / disabled? */
+static int feature_skip(State *s, int system_inode)
+{
+	switch (system_inode) {
+		case USER_QUOTA_SYSTEM_INODE:
+		case LOCAL_USER_QUOTA_SYSTEM_INODE:
+			return !(s->feature_flags.opt_ro_compat &
+					OCFS2_FEATURE_RO_COMPAT_USRQUOTA);
+		case GROUP_QUOTA_SYSTEM_INODE:
+		case LOCAL_GROUP_QUOTA_SYSTEM_INODE:
+			return !(s->feature_flags.opt_ro_compat &
+					OCFS2_FEATURE_RO_COMPAT_GRPQUOTA);
+		default:
+			return 0;
+	}
+}
+
 static inline uint32_t system_dir_bytes_needed(State *s)
 {
 	int each = OCFS2_DIR_REC_LEN(SYSTEM_FILE_NAME_MAX);
@@ -233,6 +256,114 @@ static inline uint32_t system_dir_bytes_needed(State *s)
 	return each * sys_blocks_needed(s->initial_slots);
 }
 
+static void format_quota_files(State *s, ocfs2_filesys *fs)
+{
+	errcode_t ret;
+	ocfs2_quota_hash *usr_hash = NULL, *grp_hash = NULL;
+
+	/* Write correct data into quota files */
+	if (!feature_skip(s, USER_QUOTA_SYSTEM_INODE)) {
+		ret = ocfs2_init_fs_quota_info(fs, USRQUOTA);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while looking up global user quota file");
+			goto error;
+		}
+		fs->qinfo[USRQUOTA].flags = 0;
+		fs->qinfo[USRQUOTA].qi_info.dqi_syncms = OCFS2_DEF_QUOTA_SYNC;
+		fs->qinfo[USRQUOTA].qi_info.dqi_bgrace = OCFS2_DEF_BLOCK_GRACE;
+		fs->qinfo[USRQUOTA].qi_info.dqi_igrace = OCFS2_DEF_INODE_GRACE;
+
+		ret = ocfs2_new_quota_hash(&usr_hash);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while creating user quota hash.");
+			goto error;
+		}
+		ret = ocfs2_init_global_quota_file(fs, USRQUOTA);
+		if (ret) {
+			com_err(s->progname, ret, "while creating global user "
+				"quota file");
+			goto error;
+		}
+		ret = ocfs2_init_local_quota_files(fs, USRQUOTA);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while initializing local user quota files");
+			goto error;
+		}
+	}
+	if (!feature_skip(s, GROUP_QUOTA_SYSTEM_INODE)) {
+		ret = ocfs2_init_fs_quota_info(fs, GRPQUOTA);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while looking up global group quota file");
+			goto error;
+		}
+		fs->qinfo[GRPQUOTA].flags = 0;
+		fs->qinfo[GRPQUOTA].qi_info.dqi_syncms = OCFS2_DEF_QUOTA_SYNC;
+		fs->qinfo[GRPQUOTA].qi_info.dqi_bgrace = OCFS2_DEF_BLOCK_GRACE;
+		fs->qinfo[GRPQUOTA].qi_info.dqi_igrace = OCFS2_DEF_INODE_GRACE;
+		ret = ocfs2_new_quota_hash(&grp_hash);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while creating group quota hash.");
+			goto error;
+		}
+		ret = ocfs2_init_global_quota_file(fs, GRPQUOTA);
+		if (ret) {
+			com_err(s->progname, ret, "while creating global group "
+				"quota file");
+			goto error;
+		}
+
+		ret = ocfs2_init_local_quota_files(fs, GRPQUOTA);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while initializing local group quota files");
+			goto error;
+		}
+	}
+
+	ret = ocfs2_compute_quota_usage(fs, usr_hash, grp_hash);
+	if (ret) {
+		com_err(s->progname, ret, "while computing quota usage");
+		goto error;
+	}
+	if (usr_hash) {
+		ret = ocfs2_write_release_dquots(fs, USRQUOTA, usr_hash);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while writing user quota usage");
+			goto error;
+		}
+		ret = ocfs2_free_quota_hash(usr_hash);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while releasing user quota hash");
+			goto error;
+		}
+	}
+	if (grp_hash) {
+		ret = ocfs2_write_release_dquots(fs, GRPQUOTA, grp_hash);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while writing group quota usage");
+			goto error;
+		}
+		ret = ocfs2_free_quota_hash(grp_hash);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while releasing group quota hash");
+			goto error;
+		}
+	}
+	return;
+error:
+	clear_both_ends(s);
+	exit(1);
+}
+
 static void grow_extent_allocator(State *s, ocfs2_filesys *fs)
 {
 	errcode_t ret;
@@ -305,6 +436,26 @@ static void finish_normal_format(State *s)
 	if (!s->quiet)
 		printf("done\n");
 
+	if (!s->quiet)
+		printf("Formatting quota files: ");
+
+	format_quota_files(s, fs);
+
+	if (!s->quiet)
+		printf("done\n");
+
+	if (s->dx_dirs && !s->inline_data) {
+		/*
+		 * We want to do this after quota, but before adding
+		 * any new entries to directories.
+		 */
+		if (!s->quiet)
+			printf("Indexing system directories: ");
+		index_system_dirs(s, fs);
+		if (!s->quiet)
+			printf("done\n");
+	}
+
 	if (!s->quiet)
 		printf("Writing lost+found: ");
 	create_lost_found_dir(s, fs);
@@ -466,7 +617,8 @@ main(int argc, char **argv)
 	add_entry_to_directory(s, root_dir, "..", root_dir_rec.fe_off, OCFS2_FT_DIR);
 
 	need = system_dir_bytes_needed(s);
-	if (!s->inline_data || need > ocfs2_max_inline_data(s->blocksize)) {
+	if (!s->inline_data ||
+	    need > ocfs2_max_inline_data_with_xattr(s->blocksize, NULL)) {
 		need = system_dir_blocks_needed(s) << s->blocksize_bits;
 		alloc_bytes_from_bitmap(s, need, s->global_bm,
 					&system_dir_rec.extent_off,
@@ -483,6 +635,8 @@ main(int argc, char **argv)
 	for (i = 0; i < NUM_SYSTEM_INODES; i++) {
 		if (hb_dev_skip(s, i))
 			continue;
+		if (feature_skip(s, i))
+			continue;
 
 		num = (system_files[i].global) ? 1 : s->initial_slots;
 		for (j = 0; j < num; j++) {
@@ -541,6 +695,8 @@ main(int argc, char **argv)
 	for (i = 0; i < NUM_SYSTEM_INODES; i++) {
 		if (hb_dev_skip(s, i))
 			continue;
+		if (feature_skip(s, i))
+			continue;
 
 		num = system_files[i].global ? 1 : s->initial_slots;
 		for (j = 0; j < num; j++) {
@@ -598,20 +754,20 @@ main(int argc, char **argv)
 
 static void
 parse_fs_type_opts(char *progname, const char *typestr,
-		   enum ocfs2_fs_types *fs_type)
+		   enum ocfs2_mkfs_types *fs_type)
 {
 	int i;
 
-	*fs_type = FS_DEFAULT;
+	*fs_type = OCFS2_MKFSTYPE_DEFAULT;
 
-	for(i = 0; ocfs2_fs_types_table[i].ft_str; i++) {
-		if (strcmp(typestr, ocfs2_fs_types_table[i].ft_str) == 0) {
-			*fs_type = ocfs2_fs_types_table[i].ft_type;
+	for(i = 0; ocfs2_mkfs_types_table[i].ft_str; i++) {
+		if (strcmp(typestr, ocfs2_mkfs_types_table[i].ft_str) == 0) {
+			*fs_type = ocfs2_mkfs_types_table[i].ft_type;
 			break;
 		}
 	}
 
-	if (*fs_type == FS_DEFAULT) {
+	if (*fs_type == OCFS2_MKFSTYPE_DEFAULT) {
 		com_err(progname, 0, "Bad fs type option specified.");
 		exit(1);
 	}
@@ -637,7 +793,7 @@ get_state(int argc, char **argv)
 	uint64_t val;
 	uint64_t journal_size_in_bytes = 0;
 	int journal64 = 0;
-	enum ocfs2_fs_types fs_type = FS_DEFAULT;
+	enum ocfs2_mkfs_types fs_type = OCFS2_MKFSTYPE_DEFAULT;
 	int mount = -1;
 	int no_backup_super = -1;
 	enum ocfs2_feature_levels level = OCFS2_FEATURE_LEVEL_DEFAULT;
@@ -861,6 +1017,8 @@ get_state(int argc, char **argv)
 	if ((optind == argc) && !show_version)
 		usage(progname);
 
+	srand48(time(NULL));
+
 	device_name = argv[optind];
 	optind++;
 
@@ -917,6 +1075,7 @@ get_state(int argc, char **argv)
 	s->fs_type = fs_type;
 
 	ret = ocfs2_merge_feature_flags_with_level(&s->feature_flags,
+						   fs_type,
 						   level,
 						   &feature_flags,
 						   &reverse_flags);
@@ -939,6 +1098,10 @@ get_state(int argc, char **argv)
 		s->no_backup_super = 0;
 	else
 		s->no_backup_super = 1;
+	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_INDEXED_DIRS)
+		s->dx_dirs = 1;
+	else
+		s->dx_dirs = 0;
 
 
 	/* Here if the user set these flags explicitly, we will use them and
@@ -1140,6 +1303,15 @@ static unsigned int journal_size_mail(State *s)
 	return 65536;
 }
 
+static unsigned int journal_size_vmstore(State *s)
+{
+	if (s->volume_size_in_blocks < 262144)
+		return 8192;
+	else if (s->volume_size_in_blocks < 524288)
+		return 16384;
+	return 32768;
+}
+
 /* stolen from e2fsprogs, modified to fit ocfs2 patterns */
 static uint64_t figure_journal_size(uint64_t size, State *s)
 {
@@ -1169,12 +1341,15 @@ static uint64_t figure_journal_size(uint64_t size, State *s)
 	}
 
 	switch (s->fs_type) {
-	case FS_DATAFILES:
+	case OCFS2_MKFSTYPE_DATAFILES:
 		j_blocks = journal_size_datafiles();
 		break;
-	case FS_MAIL:
+	case OCFS2_MKFSTYPE_MAIL:
 		j_blocks = journal_size_mail(s);
 		break;
+	case OCFS2_MKFSTYPE_VMSTORE:
+		j_blocks = journal_size_vmstore(s);
+		break;
 	default:
 		j_blocks = journal_size_default(s);
 		break;
@@ -1237,10 +1412,11 @@ static uint32_t figure_extent_alloc_size(State *s)
 		return 0;
 
 	switch (s->fs_type) {
-	case FS_DATAFILES:
+	case OCFS2_MKFSTYPE_DATAFILES:
+	case OCFS2_MKFSTYPE_VMSTORE:
 		target_percent = 0.3;
 		break;
-	case FS_MAIL:
+	case OCFS2_MKFSTYPE_MAIL:
 	default:
 		target_percent = 0.1;
 		break;
@@ -1397,7 +1573,8 @@ fill_defaults(State *s)
 
 	if (!s->cluster_size) {
 		switch (s->fs_type) {
-		case FS_DATAFILES:
+		case OCFS2_MKFSTYPE_DATAFILES:
+		case OCFS2_MKFSTYPE_VMSTORE:
 			s->cluster_size = cluster_size_datafiles(s);
 			break;
 		default:
@@ -1534,7 +1711,8 @@ initialize_alloc_group(State *s, const char *name,
 
 	strcpy((char *)group->gd->bg_signature, OCFS2_GROUP_DESC_SIGNATURE);
 	group->gd->bg_generation = s->vol_generation;
-	group->gd->bg_size = (uint32_t)ocfs2_group_bitmap_size(s->blocksize);
+	group->gd->bg_size =
+			(uint32_t)ocfs2_group_bitmap_size(s->blocksize, 0, 0);
 	group->gd->bg_bits = cpg * bpc;
 	group->gd->bg_chain = chain;
 	group->gd->bg_parent_dinode = alloc_inode->fe_off >> 
@@ -2015,6 +2193,24 @@ static void mkfs_swap_inode_from_cpu(State *s, struct ocfs2_dinode *di)
 	ocfs2_swap_inode_from_cpu(&fake_fs, di);
 }
 
+static void mkfs_swap_group_desc_from_cpu(State *s, struct ocfs2_group_desc *gd)
+{
+	ocfs2_filesys fake_fs;
+	char super_buf[OCFS2_MAX_BLOCKSIZE];
+
+	fill_fake_fs(s, &fake_fs, super_buf);
+	ocfs2_swap_group_desc_from_cpu(&fake_fs, gd);
+}
+
+static void mkfs_swap_group_desc_to_cpu(State *s, struct ocfs2_group_desc *gd)
+{
+	ocfs2_filesys fake_fs;
+	char super_buf[OCFS2_MAX_BLOCKSIZE];
+
+	fill_fake_fs(s, &fake_fs, super_buf);
+	ocfs2_swap_group_desc_to_cpu(&fake_fs, gd);
+}
+
 static void
 format_superblock(State *s, SystemFileDiskRecord *rec,
 		  SystemFileDiskRecord *root_rec, SystemFileDiskRecord *sys_rec)
@@ -2082,11 +2278,9 @@ format_superblock(State *s, SystemFileDiskRecord *rec,
 	 */
 	s->feature_flags.opt_compat &= ~OCFS2_FEATURE_COMPAT_BACKUP_SB;
 
-	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_XATTR) {
+	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_XATTR)
 		di->id2.i_super.s_xattr_inline_size =
 						OCFS2_MIN_XATTR_INLINE_SIZE;
-		di->id2.i_super.s_uuid_hash = ocfs2_xattr_uuid_hash(s->uuid);
-	}
 
 	di->id2.i_super.s_feature_incompat = s->feature_flags.opt_incompat;
 	di->id2.i_super.s_feature_compat = s->feature_flags.opt_compat;
@@ -2095,6 +2289,17 @@ format_superblock(State *s, SystemFileDiskRecord *rec,
 	strcpy((char *)di->id2.i_super.s_label, s->vol_label);
 	memcpy(di->id2.i_super.s_uuid, s->uuid, 16);
 
+	/* s_uuid_hash is also used by Indexed Dirs */
+	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_XATTR ||
+	    s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_INDEXED_DIRS)
+		di->id2.i_super.s_uuid_hash = ocfs2_xattr_uuid_hash(s->uuid);
+
+	if (s->feature_flags.opt_incompat & OCFS2_FEATURE_INCOMPAT_INDEXED_DIRS) {
+		di->id2.i_super.s_dx_seed[0] = mrand48();
+		di->id2.i_super.s_dx_seed[1] = mrand48();
+		di->id2.i_super.s_dx_seed[2] = mrand48();
+	}
+
 	mkfs_swap_inode_from_cpu(s, di);
 	mkfs_compute_meta_ecc(s, di, &di->i_check);
 	do_pwrite(s, di, s->blocksize, super_off);
@@ -2180,7 +2385,7 @@ format_file(State *s, SystemFileDiskRecord *rec)
 		di->id2.i_chain.cl_count = 
 			ocfs2_chain_recs_per_inode(s->blocksize);
 		di->id2.i_chain.cl_cpg =
-				 ocfs2_group_bitmap_size(s->blocksize) * 8;
+			 ocfs2_group_bitmap_size(s->blocksize, 0, 0) * 8;
 		di->id2.i_chain.cl_bpc = 1;
 		if (s->nr_cluster_groups > 
 		    ocfs2_chain_recs_per_inode(s->blocksize)) {
@@ -2243,21 +2448,25 @@ format_file(State *s, SystemFileDiskRecord *rec)
 			(struct ocfs2_dir_entry *)(dir->buf + dir->last_off);
 		int dir_len = dir->last_off + OCFS2_DIR_REC_LEN(de->name_len);
 
-		if (dir_len > ocfs2_max_inline_data(s->blocksize)) {
+		if (dir_len >
+		    ocfs2_max_inline_data_with_xattr(s->blocksize, di)) {
 			com_err(s->progname, 0,
 				"Inline a dir which shouldn't be inline.\n");
 			clear_both_ends(s);
 			exit(1);
 		}
 		de->rec_len -= s->blocksize -
-			       ocfs2_max_inline_data(s->blocksize);
+			       ocfs2_max_inline_data_with_xattr(s->blocksize,
+								di);
 		memset(&di->id2, 0,
 		       s->blocksize - offsetof(struct ocfs2_dinode, id2));
 
-		di->id2.i_data.id_count = ocfs2_max_inline_data(s->blocksize);
+		di->id2.i_data.id_count =
+			ocfs2_max_inline_data_with_xattr(s->blocksize, di);
 		memcpy(di->id2.i_data.id_data, dir->buf, dir_len);
 		di->i_dyn_features |= OCFS2_INLINE_DATA_FL;
-		di->i_size = ocfs2_max_inline_data(s->blocksize);
+		di->i_size = ocfs2_max_inline_data_with_xattr(s->blocksize,
+							      di);
 	}
 
 write_out:
@@ -2306,7 +2515,7 @@ write_bitmap_data(State *s, AllocBitmap *bitmap)
 		gd->bg_parent_dinode = parent_blkno;
 		memcpy(buf, gd, s->blocksize);
 		gd_buf = (struct ocfs2_group_desc *)buf;
-		ocfs2_swap_group_desc(gd_buf);
+		mkfs_swap_group_desc_from_cpu(s, gd_buf);
 		mkfs_compute_meta_ecc(s, buf, &gd_buf->bg_check);
 		do_pwrite(s, buf, s->cluster_size,
 			  gd->bg_blkno << s->blocksize_bits);
@@ -2318,10 +2527,10 @@ static void
 write_group_data(State *s, AllocGroup *group)
 {
 	uint64_t blkno = group->gd->bg_blkno;
-	ocfs2_swap_group_desc(group->gd);
+	mkfs_swap_group_desc_from_cpu(s, group->gd);
 	mkfs_compute_meta_ecc(s, group->gd, &group->gd->bg_check);
 	do_pwrite(s, group->gd, s->blocksize, blkno << s->blocksize_bits);
-	ocfs2_swap_group_desc(group->gd);
+	mkfs_swap_group_desc_to_cpu(s, group->gd);
 }
 
 static void mkfs_swap_dir(State *s, DirData *dir,
@@ -2500,6 +2709,9 @@ init_record(State *s, SystemFileDiskRecord *rec, int type, int mode)
 	case SFI_TRUNCATE_LOG:
 		rec->flags |= OCFS2_DEALLOC_FL;
 		break;
+	case SFI_QUOTA:
+		rec->flags |= OCFS2_QUOTA_FL;
+		break;
 	case SFI_OTHER:
 		break;
 	}
@@ -2526,11 +2738,11 @@ print_state(State *s)
 
 	ocfs2_snprint_feature_flags(buf, PATH_MAX, &s->feature_flags);
 
-	if (s->fs_type != FS_DEFAULT) {
-		for(i = 0; ocfs2_fs_types_table[i].ft_str; i++) {
-			if (ocfs2_fs_types_table[i].ft_type == s->fs_type) {
+	if (s->fs_type != OCFS2_MKFSTYPE_DEFAULT) {
+		for(i = 0; ocfs2_mkfs_types_table[i].ft_str; i++) {
+			if (ocfs2_mkfs_types_table[i].ft_type == s->fs_type) {
 				printf("Filesystem Type of %s\n",
-				       ocfs2_fs_types_table[i].ft_str);
+				       ocfs2_mkfs_types_table[i].ft_str);
 				break;
 			}
 		}
@@ -2575,6 +2787,44 @@ clear_both_ends(State *s)
 	return ;
 }
 
+static void index_system_dirs(State *s, ocfs2_filesys *fs)
+{
+	errcode_t ret;
+	int i, num_slots = OCFS2_RAW_SB(fs->fs_super)->s_max_slots;
+	uint64_t orphan_dir_blkno;
+
+
+	/* Start with the root directory */
+	ret = ocfs2_dx_dir_build(fs, fs->fs_root_blkno);
+	if (ret) {
+		com_err(s->progname, ret, "while indexing root directory");
+		goto bail;
+	}
+
+	for (i = 0; i < num_slots; i++) {
+		ret = ocfs2_lookup_system_inode(fs, ORPHAN_DIR_SYSTEM_INODE,
+						i, &orphan_dir_blkno);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while looking up orphan dir %d for indexing",
+				i);
+			goto bail;
+		}
+
+		ret = ocfs2_dx_dir_build(fs, orphan_dir_blkno);
+		if (ret) {
+			com_err(s->progname, ret, "while indexing root directory");
+			goto bail;
+		}
+	}
+
+	return;
+
+bail:
+	clear_both_ends(s);
+	exit(1);
+}
+
 static void create_lost_found_dir(State *s, ocfs2_filesys *fs)
 {
 	errcode_t ret;
