@@ -96,15 +96,10 @@ enum {
 	SFI_HEARTBEAT,
 	SFI_CHAIN,
 	SFI_TRUNCATE_LOG,
+	SFI_QUOTA,
 	SFI_OTHER
 };
 
-enum ocfs2_fs_types {
-	FS_DEFAULT,
-	FS_DATAFILES,
-	FS_MAIL
-};
-
 typedef struct _SystemFileInfo SystemFileInfo;
 
 struct _SystemFileInfo {
@@ -193,6 +188,7 @@ struct _State {
 	int mount;
 	int no_backup_super;
 	int inline_data;
+	int dx_dirs;
 	int dry_run;
 
 	uint32_t blocksize;
@@ -238,7 +234,7 @@ struct _State {
 
 	ocfs2_fs_options feature_flags;
 
-	enum ocfs2_fs_types fs_type;
+	enum ocfs2_mkfs_types fs_type;
 };
 
 int ocfs2_fill_cluster_information(State *s);
